//
// Author: Benjamin Langmann <b.langmann@gmx.de>, (C) 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "stdafx.h" 
#include "lmjpeg_encoder.h"
#include "lmjpeg_io.h"
#include "huffman.h"
#include "helpers.h"
#include <math.h>

#ifndef VFW_WIN
#include <zlib.h>
#else
#include "zlib\zlib.h"
#endif

//#define GEN_TREE

/** allocates and initializes a CODEC struct.
	@return the allocated and initialized CODEC struct
*/
CODEC* setup_codec() {
	int i;
	CODEC *codec=Malloc(sizeof(CODEC));
	codec->config=Malloc(sizeof(lmjpeg_fileheader));

	codec->config->marker=LMJPEG_IO_FILEHEADER;
	codec->config->width=0;
	codec->config->height=0;
	codec->config->topdown=1;
	codec->config->time_base_num=1;
	codec->config->time_base_den=1;
	codec->config->keyframe_intervall=100;
	codec->config->dyn_keyframes=0;
	codec->config->quality_mode=LMJPEG_MODE_SUMDIFF;
	codec->config->quality_num=3;
	codec->config->quality_diff=10;
	codec->config->jpeg_quality=70;
	codec->config->pix_format=LMJPEG_YV12;

	lmjpeg_calc_QM(codec);
	for (i=0;i<64;i++) {
		codec->config->ratings[i]=10;
	}
	/*codec->config->ratings[0]=10;
	codec->config->ratings[1]=10;
	codec->config->ratings[8]=10;
	codec->config->ratings[2]=10;
	codec->config->ratings[9]=10;
	codec->config->ratings[16]=10;*/
	codec->config->compression=LMJPEG_IO_RLE;

	codec->frame_num=0;
	codec->save=0;
	codec->keyframe_cnt=0;
	codec->frame_type=LMJPEG_IO_KEYFRAME;
	codec->last_frame=NULL;
	codec->lmjpeg_file=NULL;
	codec->written_frame_bytes=0;
	codec->last_frame_num=0;
	codec->file_size=0;
	codec->want_topdown=1;
	codec->in_out_pix_format=LMJPEG_BGR;

	codec->field_widths_int[0]=0;
	codec->field_widths_int[1]=0;
	codec->field_heights_int[0]=0;
	codec->field_heights_int[1]=0;
#ifdef GEN_TREE
	huffman_gen_tree(NULL,0);
#endif
	return codec;
}

/** frees a CODEC struct.
	@param co the CODEC struct to be freed
*/
void end_codec(CODEC *co) {
	Free(co->config);
	co->config=NULL;
	Free(co);
#ifdef GEN_TREE
	huffman_gen_tree(NULL,-1);
#endif
}

/** checks if the values of a configuration are valid.
	@param config the configuration to be checked
	@return 0 if a value is not valid
*/
int check_config(lmjpeg_fileheader *config) {
	if (config==NULL) return 0;
	if (config->marker!=LMJPEG_IO_FILEHEADER) return 0;
	if (config->time_base_den<0) return 0;
	if (config->time_base_num<0) return 0;
	if (config->keyframe_intervall<0) return 0;
	if (config->quality_mode!=LMJPEG_MODE_NUMPIX && config->quality_mode!=LMJPEG_MODE_SUMDIFF) return 0;
	if (config->quality_num<0) return 0;
	if (config->quality_diff<0) return 0;
	if (config->jpeg_quality<0) return 0;
	if (config->jpeg_quality>132) return 0;
	return 1;
}

#define QM_MIN 1
#define QM_MAX 63

#define QM_0 5.0
#define QM_99  0.1

/** generates the quatisation matrix to be used.
	@param codec a pointer to the CODEC struct
*/
void lmjpeg_calc_QM(CODEC *codec) {
	int i;
	double r;
	if (codec->config->jpeg_quality>=0 && codec->config->jpeg_quality<=100) {
		// scales linear from 99 - 0 --> QM_99 - QM_0
		r=QM_0+(((QM_99-QM_0)/99.0)*(double) codec->config->jpeg_quality);
		printf("calc QM: JPEG: %d r: %f\n",codec->config->jpeg_quality,r);
		for (i=0;i<64;i++) {
			codec->config->myqm[i]=0.5+(double) QM[i] * r;
			if (codec->config->myqm[i]<QM_MIN) codec->config->myqm[i]=QM_MIN;
			if (codec->config->myqm[i]>QM_MAX) codec->config->myqm[i]=QM_MAX;
		}
	}
	
	// nothing to be done for mpeg_quant and manual QM
}

/** wrapper function for the lossless compression methodes.
	@param codec a pointer to the CODEC struct
	@param data pointer to the data to be compressed
	@param size size of the data to be compressed
	@return size of the compressed data in bytes
*/
int lmjpeg_compress(CODEC *codec, uint16_t *data, int size) {
	uint16_t *tmp=(uint16_t*) Malloc(5*size*sizeof(uint16_t)/4);

	if (codec->config->compression==LMJPEG_IO_RLE) {
		if (codec->config->jpeg_quality!=100) {
			size=rle_encode_12(data, size, tmp);
#ifdef GEN_TREE
			huffman_gen_tree(tmp, size);
#endif
			size=huffman_encode_12(tmp, size, (uint8_t*) data);
		}
		else {	// here size is in bytes
			//size=rle_encode((uint8_t*) data, size, (uint8_t*) tmp);
			//size=huffman_encode((uint8_t*) tmp, size, (uint8_t*) data);
			memcpy(tmp, data, size);
			size=rle_encode((uint8_t*) tmp, size, (uint8_t*) data);
		}
	}
	else {
		if (codec->config->jpeg_quality!=100) {
			memcpy(tmp, data, 2*size);
			size=zlib_encode((uint8_t*) tmp, 2*size, (uint8_t*) data);
		}
		else { // here size is in bytes
			memcpy(tmp, data, size);
			size=zlib_encode((uint8_t*) tmp, size, (uint8_t*) data);
		}
	}
	Free(tmp);
	return size;
}

/** wrapper function for the lossless decompression methodes.
	@param codec a pointer to the CODEC struct
	@param data pointer to the data to be decompressed
	@param size size of the data to be decompressed in bytes
	@param to pointer to the memory to contain the decompressed data
	@param tosize size of the memory to contain the decompressed data in uint16_t
	@return size of the decompressed data in bytes
*/
int lmjpeg_decompress(CODEC *codec, uint8_t *data, int size, uint16_t *to, int tosize) {
	uint16_t *tmp;
	
	if (codec->config->compression==LMJPEG_IO_RLE) {
		if (codec->config->jpeg_quality!=100) {
			tmp=(uint16_t*) Malloc(tosize*sizeof(uint16_t));
			size=huffman_decode_12(data, size, tmp, tosize);
			tosize=rle_decode_12(tmp, size, to, tosize);
			Free(tmp);
			return tosize*2;
		}
		else {
			//tmp=(uint16_t*) Malloc(tosize*sizeof(uint16_t));
			//size=huffman_decode(data, size, (uint8_t*) tmp, 2*tosize);
			//tosize=rle_decode((uint8_t*) tmp, size, (uint8_t*) to, tosize);
			//Free(tmp);
			tosize=rle_decode(data, size, (uint8_t*) to, tosize);
			return tosize;
		}
	}
	else {
		if (codec->config->jpeg_quality!=100)
			return zlib_decode(data, size, (uint8_t*) to, 2*tosize);
		else 
			return zlib_decode(data, size, (uint8_t*) to, 2*tosize);
	}
}

#define CUT_NUMBER 128

/** performs a run-length-encoding with 128 treated as zero.
	@param data pointer to the data to be encoded
	@param size size of the data to be encoded
	@param to pointer to the memory to contain the encoded data
	@return size of the encoded data
*/
int rle_encode(uint8_t *data, int size, uint8_t *to) {
	int i, last_was_zero=0, num_z=0, s=0;

	for (i=0;i<size;i++) {
		if (last_was_zero==0) {
			to[s++]=data[i];
			if (data[i]==CUT_NUMBER) last_was_zero=1;
		}
		else {
			if (data[i]==CUT_NUMBER) num_z++; 
			else {
				to[s++]=num_z+1;
				num_z=0;
				to[s++]=data[i];
				last_was_zero=0;
			}
		}
		if(num_z>=254) {
			to[s++]=num_z+1;
			num_z=0;
			last_was_zero=0;
		}
	}
	if(last_was_zero>0) to[s++]=num_z+1;
	return s;
}

#define CUT_NUMBER_12 2048

/** performs a run-length-encoding with 2048 treated as zero.
	@param data pointer to the data to be encoded
	@param size size of the data to be encoded
	@param to pointer to the memory to contain the encoded data
	@return size of the encoded data
*/
int rle_encode_12(uint16_t *data, int size, uint16_t *to) {
	int i, last_was_zero=0, num_z=0, s=0;

	for (i=0;i<size;i++) {
		if (last_was_zero==0) {
			to[s++]=data[i];
			if (data[i]==CUT_NUMBER_12) last_was_zero=1;
		}
		else {
			if (data[i]==CUT_NUMBER_12) num_z++; 
			else {
				to[s++]=num_z+1;
				num_z=0;
				to[s++]=data[i];
				last_was_zero=0;
			}
		}
		if(num_z>=4094) {
			to[s++]=num_z+1;
			num_z=0;
			last_was_zero=0;
		}
	}
	if(last_was_zero>0) to[s++]=num_z+1;
	return s;
}

/** performs a run-length-decoding with 128 treated as zero.
	@param data pointer to the data to be decoded
	@param size size of the data to be decoded
	@param to pointer to the memory to contain the decoded data
	@param tosize size of the memory to contain the decoded data
	@return size of the decoded data
*/
int rle_decode(uint8_t *data, int size, uint8_t *to, int tosize) {
	int i,j, pos=0, last_was_zero=0;
	for(i=0;i<size;i++) {
		if (last_was_zero==0) {
			to[pos++]=data[i];
			if (data[i]==CUT_NUMBER) last_was_zero=1;
		}
		else {
			last_was_zero=0;
			for (j=0;j<data[i]-1;j++) to[pos+j]=CUT_NUMBER;
			pos+=data[i]-1;
		}
	}
	return pos;
}

/** performs a run-length-decoding with 2048 treated as zero.
	@param data pointer to the data to be decoded
	@param size size of the data to be decoded
	@param to pointer to the memory to contain the decoded data
	@param tosize size of the memory to contain the decoded data
	@return size of the decoded data
*/
int rle_decode_12(uint16_t *data, int size, uint16_t *to, int tosize) {
	int i,j, pos=0, last_was_zero=0;
	for(i=0;i<size;i++) {
		if (last_was_zero==0) {
			to[pos++]=data[i];
			if (data[i]==CUT_NUMBER_12) last_was_zero=1;
		}
		else {
			last_was_zero=0;
			for (j=0;j<data[i]-1;j++) to[pos+j]=CUT_NUMBER_12;
			pos+=data[i]-1;
		}
	}
	return pos;
}

/** performs a compression using the zlib library.
	@param data pointer to the data to be compressed
	@param size size of the data to be compressed
	@param to pointer to the memory to contain the compressed data
	@return size of the compressed data
*/
int zlib_encode(uint8_t *data, int size, uint8_t *to) {
	unsigned long dsize=size;
	compress2(to, &dsize, data, size, Z_BEST_COMPRESSION); // eg. Z_BEST_SPEED,Z_BEST_COMPRESSION
	return dsize;	
}

/** performs a decompression using the zlib library.
	@param data pointer to the data to be decompressed
	@param size size of the data to be decompressed
	@param to pointer to the memory to contain the decompressed data
	@param tosize size of the memory to contain the decompressed data
	@return size of the decompressed data
*/
int zlib_decode(uint8_t *data, int size, uint8_t *to, int tosize) {
	unsigned long dsize=tosize;
	uncompress(to,&dsize, data, size);
	return dsize;
}

#ifndef VFW_WIN

/** writes chars to the avi file.
	@param codec a pointer to the CODEC struct
	@param data pointer to the data to be written
	@param size size of the data
	@return size of the data written or -1 on error
*/
int lmjpeg_write_char(CODEC *codec, char* data, int size) {
	int ret=fwrite(data,1,size,codec->lmjpeg_file);
	if (ret<size) {
		printf("Error writing!\n");
		return -1;
	}
	return ret;
}

/** writes an int to the avi file.
	@param codec a pointer to the CODEC struct
	@param data integer to be written
	@return size of the data written or -1 on error
*/
int lmjpeg_write_int(CODEC *codec, int data) {
	int ret=fwrite(&data,1,4,codec->lmjpeg_file);
	if (ret<4) {
		printf("Error writing!\n");
		return -1;
	}
	return ret;
}

/** writes dta to the avi file.
	@param codec a pointer to the CODEC struct
	@param data pointer to the data to be written
	@param size size of the data
	@return size of the data written or -1 on error
*/
int lmjpeg_write_data(CODEC *codec, void *data, int size) {
	int ret=fwrite(data,1,size,codec->lmjpeg_file);
	if (ret<size) {
		printf("Error writing!\n");
		return -1;
	}
	return ret;
}

/** writes the avi header to the file.
	@param codec a pointer to the CODEC struct
	@return -1 on error
*/
int lmjpeg_write_header(CODEC *codec) {
	int list_len=8+4+sizeof(lmjpeg_AVIStreamHeader)+4+4+sizeof(lmjpeg_BITMAPINFOHEADER);
	//int avi_list_len=8+4+sizeof(lmjpeg_MainAVIHeader)+4+list_len;
	// init headers
	lmjpeg_MainAVIHeader aviheader;
	lmjpeg_AVIStreamHeader streamheader;
	lmjpeg_BITMAPINFOHEADER streamformatheader;
	
	codec->config->marker=LMJPEG_IO_FILEHEADER;
	aviheader.dwMicroSecPerFrame=1000000.0*((double) codec->config->time_base_num/(double) codec->config->time_base_den);
	// frame display rate (or 0L)
	aviheader.dwMaxBytesPerSec=0;		// max. transfer rate
	aviheader.dwPaddingGranularity=AVI_HEADERSIZE;	// pad to multiples of this size; normally 2K.
	aviheader.dwFlags=AVIF_HASINDEX;
	aviheader.dwTotalFrames=codec->last_frame_num+1;
	aviheader.dwInitialFrames=0;
	aviheader.dwStreams=1;
	aviheader.dwSuggestedBufferSize=0;
	aviheader.dwWidth=codec->config->width;
	aviheader.dwHeight=codec->config->height;
	aviheader.dwReserved[0]=0;aviheader.dwReserved[1]=0;aviheader.dwReserved[2]=0;aviheader.dwReserved[3]=0;
	
	memcpy(&streamheader.fccType,"vids",4);
	memcpy(&streamheader.fccHandler,"lmjp",4);
	streamheader.dwFlags=0;	/* Contains AVITF_* flags */
	streamheader.wPriority=0;
	streamheader.wLanguage=0;
	streamheader.dwInitialFrames=0;
	streamheader.dwScale=codec->config->time_base_num;	
	streamheader.dwRate=codec->config->time_base_den;	/* dwRate / dwScale == samples/second */
	streamheader.dwStart=0;
	streamheader.dwLength=codec->last_frame_num+1; /* In units above... */
	streamheader.dwSuggestedBufferSize=0;
	streamheader.dwQuality=10000;
	streamheader.dwSampleSize=0;
	streamheader.rcFrame[0]=codec->config->width;
	streamheader.rcFrame[1]=codec->config->height;
	
	streamformatheader.biSize=sizeof(lmjpeg_BITMAPINFOHEADER);
	streamformatheader.biWidth=codec->config->width;
	streamformatheader.biHeight=codec->config->height;
	streamformatheader.biPlanes=1;
	streamformatheader.biBitCount=24;
	memcpy(&streamformatheader.biCompression,"LMJP",4);
	streamformatheader.biSizeImage=codec->config->width*codec->config->height*3;
	streamformatheader.biXPelsPerMeter=0;
	streamformatheader.biYPelsPerMeter=0;
	streamformatheader.biClrUsed=0;
	streamformatheader.biClrImportant=0;

	// write to file
	fseek(codec->lmjpeg_file,0,SEEK_SET);
	lmjpeg_write_char(codec, "RIFF",4);
	lmjpeg_write_int(codec, codec->file_size-8);	// filesize
	lmjpeg_write_char(codec, "AVI LIST",8);
	lmjpeg_write_int(codec, AVI_HEADERSIZE-20);			// List length
	lmjpeg_write_char(codec, "hdrlavih",8);
	lmjpeg_write_int(codec, sizeof(lmjpeg_MainAVIHeader)); 	// avi header size
	lmjpeg_write_data(codec, &aviheader, sizeof(lmjpeg_MainAVIHeader));	// avi header
	lmjpeg_write_char(codec, "LIST",4);				// stream list
	lmjpeg_write_int(codec, list_len);				// List length
	
	lmjpeg_write_char(codec, "strlstrh",8);			// stream 1
	lmjpeg_write_int(codec, sizeof(lmjpeg_AVIStreamHeader));			// stream 1 header length
	lmjpeg_write_data(codec, &streamheader, sizeof(lmjpeg_AVIStreamHeader));	// stream 1 header
	lmjpeg_write_char(codec, "strf",4);				// stream 1 format
	lmjpeg_write_int(codec, sizeof(lmjpeg_BITMAPINFOHEADER));			// stream 1 format length
	lmjpeg_write_data(codec, &streamformatheader, sizeof(lmjpeg_BITMAPINFOHEADER));	// stream 1 format
	lmjpeg_write_char(codec, "JUNK",4);				// fill up to 2kb (allready done!)
	lmjpeg_write_int(codec, AVI_HEADERSIZE-ftell(codec->lmjpeg_file)-4);
	fseek(codec->lmjpeg_file,AVI_HEADERSIZE,SEEK_SET);
	lmjpeg_write_char(codec, "LIST",4);
	lmjpeg_write_int(codec, codec->written_frame_bytes+4);	// +4 for "movi"
	lmjpeg_write_char(codec, "movi",4);
	return 0;
}

/** initializes the writer.
	@param codec a pointer to the CODEC struct
	@return 0 on success, -1 on error
*/
int lmjpeg_writer_init(CODEC *codec) {
	int ret;
	codec->lmjpeg_file=fopen(codec->filename,"wb");
	if (codec->lmjpeg_file==NULL) {
		printf("Error opening %s!\n",codec->filename);
		return -1;
	}
	lmjpeg_write_char(codec, "RIFF",4);
	for (ret=0;ret<AVI_HEADERSIZE/4+2;ret++) lmjpeg_write_int(codec, 0); // -1 + space for LIST,SIZE,movi
	codec->written_frame_bytes=0;
	return 0;
}

/** writes the index to the avi file
	@param codec a pointer to the CODEC struct
	@return 0 on success, -1 on error
*/
int lmjpeg_write_index(CODEC *codec) {
	lmjpeg_AVIINDEXENTRY idx;
	lmjpeg_frameheader fheader;
	FILE *reader;
	int written=0;
	long index_start, index_end;
	
	// set up writer
	index_start=ftell(codec->lmjpeg_file);
	fseek(codec->lmjpeg_file,8,SEEK_CUR);
	
	// set up reader
	reader=fopen(codec->filename,"rb");
	if (reader==NULL) {
		printf("Error opening %s!\n",codec->filename);
		return -1;
	}
	fseek(reader,AVI_HEADERSIZE+12,SEEK_SET);
	
	// read frames and write index
	while (!feof(reader)) {
		idx.dwChunkOffset=ftell(reader);
		if (idx.dwChunkOffset>=index_start) break;
		
		idx.dwChunkOffset=idx.dwChunkOffset-(AVI_HEADERSIZE+8);
		idx.ckid=lmjpeg_read_int(reader);
		idx.dwChunkLength=lmjpeg_read_int(reader);

		if (memcmp(&idx.ckid,"00dc",4)==0) {
			lmjpeg_read_data(reader,&fheader,sizeof(lmjpeg_frameheader));
			if (fheader.size%2==1) fseek(reader,fheader.size+1,SEEK_CUR);
			else fseek(reader,fheader.size,SEEK_CUR);
			
			if (fheader.frame_num==0) { 
				fseek(reader,sizeof(lmjpeg_fileheader),SEEK_CUR);
			}
			//if (codec->config->keyframe_intervall > 0 && fheader.frame_num%codec->config->keyframe_intervall==0)
			if ((fheader.marker & ~LMJPEG_IO_FRAMEHEADERMASK) == LMJPEG_IO_KEYFRAME)
				idx.dwFlags=AVIIF_KEYFRAME;
			else
				idx.dwFlags=0;
			//}
		}
		else if (memcmp(&idx.ckid,"idx1",4)==0) {
			break;
		}
		else {
			break;
			fseek(reader,idx.dwChunkLength,SEEK_CUR);
			idx.dwFlags=0;
		}
		lmjpeg_write_data(codec,&idx,sizeof(lmjpeg_AVIINDEXENTRY));
		written++;
	}
	index_end=ftell(codec->lmjpeg_file);
	fseek(codec->lmjpeg_file,index_start,SEEK_SET);
	lmjpeg_write_char(codec,"idx1",4);
	lmjpeg_write_int(codec, written*sizeof(lmjpeg_AVIINDEXENTRY));
	fclose(reader);
	
	// to write file size
	codec->file_size=AVI_HEADERSIZE+12+codec->written_frame_bytes+index_end-index_start;	// +12 for LIST|size|movi
	return 0;
}

/** writes a frame to the avi file.
	@param codec a pointer to the CODEC struct
	@param framenum number of the frame
	@param img pointer to the data of the frame
	@param size size of the data of the frame
	@return -1 on error
*/
int lmjpeg_write(CODEC *codec, int framenum, uint8_t *img, int size) {
	int ret;
	if (codec->lmjpeg_file==NULL) {
		printf("Error: video not opened!\n");
		return -1;
	}
	codec->last_frame_num=framenum;
	codec->written_frame_bytes+=8+size+sizeof(lmjpeg_frameheader);
	
	lmjpeg_frameheader fheader;
	fheader.marker=LMJPEG_IO_FRAMEHEADER | codec->frame_type;
	fheader.frame_num=framenum;
	fheader.size=size;
	
	lmjpeg_write_char(codec, "00dc",4);
	if (framenum==0) {
		codec->config->marker=LMJPEG_IO_FILEHEADER;
		lmjpeg_write_int(codec, fheader.size+sizeof(lmjpeg_frameheader)+sizeof(lmjpeg_fileheader));
		lmjpeg_write_data(codec, &fheader,sizeof(lmjpeg_frameheader));
		lmjpeg_write_data(codec, codec->config,sizeof(lmjpeg_fileheader));
		codec->written_frame_bytes+=sizeof(lmjpeg_fileheader);
	}
	else {
		lmjpeg_write_int(codec, fheader.size+sizeof(lmjpeg_frameheader));
		lmjpeg_write_data(codec, &fheader,sizeof(lmjpeg_frameheader));
	}
	ret=lmjpeg_write_data(codec, img, size);
	
	// pad to 2 bytes
	if (size%2==1) {
		lmjpeg_write_char(codec, "\0",1);
		codec->written_frame_bytes++;
	}
	return ret;
}

/** writes header, index and closes the file.
	@param codec a pointer to the CODEC struct
*/
void lmjpeg_writer_close(CODEC *codec) {
	lmjpeg_write_index(codec);
	lmjpeg_write_header(codec);
	fclose(codec->lmjpeg_file);
	codec->lmjpeg_file=NULL;
}

/** read chars from the avi file.
	@param file a pointer to the avi file handle
	@param str pointer to the memory to store the read data
	@param size size of the data to be read
	@return pointer to the data read or NULL on error
*/
char* lmjpeg_read_char(FILE *file, char* str, int size) {
	int ret=fread(str,1,size,file);
	if (ret<size) {
		printf("Error reading! pos: %d\n",(int) ftell(file));
		return NULL;
	}
	return str;
}

/** read am integer from the avi file.
	@param file a pointer to the avi file handle
	@returnr read integer or -1 on error
*/
int lmjpeg_read_int(FILE *file) {
	int tmp;
	int ret=fread(&tmp,1,4,file);
	if (ret<4) {
		printf("Error reading! pos: %d\n",(int) ftell(file));
		return -1;
	}
	return tmp;
}

/** read data from the avi file.
	@param file a pointer to the avi file handle
	@param data pointer to the memory to store the read data
	@param size size of the data to be read
	@return size of the data read or -1 on error
*/
int lmjpeg_read_data(FILE *file, void *data, int size) {
	int ret=fread(data,1,size,file);
	if (ret<size) {
		printf("Error reading! pos: %d\n",(int) ftell(file));
		return -1;
	}
	return ret;
}

/** open the avi file, read headers.
	@param codec a pointer to the CODEC struct
	@return -1 on error
*/
int lmjpeg_reader_init(CODEC *codec) {
	int size;
	char str[8];
	lmjpeg_AVIStreamHeader streamheader;
	lmjpeg_BITMAPINFOHEADER streamformatheader;
	
	codec->lmjpeg_file=fopen(codec->filename,"rb");
	if (codec->lmjpeg_file==NULL) {
		printf("Error opening %s!\n",codec->filename);
		return -1;
	}
	if (strncmp(lmjpeg_read_char(codec->lmjpeg_file, str,4),"RIFF",4)!=0) {
		printf("Error: %s not a RIFF avi file! pos: %d\n",codec->filename,(int) ftell(codec->lmjpeg_file));
		return -1;
	}
	codec->file_size=lmjpeg_read_int(codec->lmjpeg_file);
	if (strncmp(lmjpeg_read_char(codec->lmjpeg_file, str,8),"AVI LIST",8)!=0) {
		printf("Error: %s not a RIFF avi file! pos: %d\n",codec->filename,(int) ftell(codec->lmjpeg_file));
		return -1;
	}
	lmjpeg_read_int(codec->lmjpeg_file);
	if (strncmp(lmjpeg_read_char(codec->lmjpeg_file, str,8),"hdrlavih",8)!=0) {
		printf("Error: %s not a RIFF avi file! pos: %d\n",codec->filename,(int) ftell(codec->lmjpeg_file));
		return -1;
	}
	size=lmjpeg_read_int(codec->lmjpeg_file);
	fseek(codec->lmjpeg_file, size, SEEK_CUR);	// AVIMainHeader
	if (strncmp(lmjpeg_read_char(codec->lmjpeg_file, str,4),"LIST",4)!=0) {
		printf("Error: %s not a RIFF avi file! pos: %d size: %d\n",codec->filename,(int) ftell(codec->lmjpeg_file),size);
		return -1;
	}
	
	// stream 1
	size=lmjpeg_read_int(codec->lmjpeg_file);
	if (strncmp(lmjpeg_read_char(codec->lmjpeg_file, str,8),"strlstrh",8)!=0) {
		printf("Error: %s not a RIFF avi file! pos: %d\n",codec->filename,(int) ftell(codec->lmjpeg_file));
		return -1;
	}
	size=lmjpeg_read_int(codec->lmjpeg_file);
	if (lmjpeg_read_data(codec->lmjpeg_file, &streamheader,size)<size) {
		printf("Error: %s read wrong stream header! pos: %d\n",codec->filename,(int) ftell(codec->lmjpeg_file));
		return -1;
	}
	if (strncmp(lmjpeg_read_char(codec->lmjpeg_file, str,4),"strf",4)!=0) {
		printf("Error: %s not a RIFF avi file! pos: %d\n",codec->filename,(int) ftell(codec->lmjpeg_file));
		return -1;
	}
	size=lmjpeg_read_int(codec->lmjpeg_file);	// format header
	if (lmjpeg_read_data(codec->lmjpeg_file, &streamformatheader,size)<size) {
		printf("Error: %s not a RIFF avi file! pos: %d\n",codec->filename,(int) ftell(codec->lmjpeg_file));
		return -1;
	}
		
	// init header:
	codec->config->width=streamformatheader.biWidth;
	if (streamformatheader.biHeight>0) {
		codec->config->height=streamformatheader.biHeight;
		codec->config->topdown=0;
	}
	else {
		codec->config->height=-streamformatheader.biHeight;
		codec->config->topdown=1;
	}
	codec->config->time_base_num=streamheader.dwScale;
	codec->config->time_base_den=streamheader.dwRate;

	// SKIP other stream headers
	while (1) {
		lmjpeg_read_char(codec->lmjpeg_file,str,4);
		if (strncmp(str,"LIST",4)!=0 && strncmp(str,"JUNK",4)!=0) {
			printf("Error: %s not a RIFF avi file! pos: %d\n",codec->filename,(int) ftell(codec->lmjpeg_file));
			return -1;
		}
		size=lmjpeg_read_int(codec->lmjpeg_file);
		if (strncmp(lmjpeg_read_char(codec->lmjpeg_file, str, 4),"movi",4)!=0) {
			fseek(codec->lmjpeg_file, size-4, SEEK_CUR);
		}
		else return 0;	// we are ok here
	}
}

/** read a frame from the avi file.
	@param codec a pointer to the CODEC struct
	@param fheader pointer to the memory to store the frame header
	@param img pointer to the data of the frame
	@return -1 on error
*/
int lmjpeg_read(CODEC *codec, lmjpeg_frameheader *fheader, uint8_t *img) {
	int ret, size_chunk, size_tmp=0;
	char str[4];
	if (codec->lmjpeg_file==NULL) {
		printf("Error: video not opened!\n");
		return -1;
	}
	//printf("start pos: %d\n",(int) ftell(lmjpeg_file));
	if (ftell(codec->lmjpeg_file)>=codec->file_size-1) return 0;	// EOF
	
	if (ftell(codec->lmjpeg_file)%2==1) fseek(codec->lmjpeg_file, 1, SEEK_CUR);
	lmjpeg_read_char(codec->lmjpeg_file, str, 4);
	//str[4]='\0';
	//printf("read %4s pos: %d\n",str,(int) ftell(lmjpeg_file));
	while (strncmp(str,"00dc",4)!=0 && strncmp(str,"00db",4)!=0) { // find right stream
			if (strncmp(str,"idx1",4)==0) return -2;
			size_chunk=lmjpeg_read_int(codec->lmjpeg_file);
			if (size_chunk<=0) return 0;	// EOF
			if (size_chunk%2==1) size_chunk+=1;
			fseek(codec->lmjpeg_file, size_chunk, SEEK_CUR);
			//printf("seek to pos: %d\n",(int) ftell(lmjpeg_file));
			lmjpeg_read_char(codec->lmjpeg_file, str, 4);
			//str[4]='\0';
			//printf("read %4s pos: %d\n",str,(int) ftell(lmjpeg_file));
	}
	
	size_chunk=lmjpeg_read_int(codec->lmjpeg_file);
	if (size_chunk<sizeof(lmjpeg_frameheader)) {
		printf("Error reading frame header! read %d bytes instead of %d\n",size_chunk,sizeof(lmjpeg_frameheader));
		return -1;
	}
	ret=lmjpeg_read_data(codec->lmjpeg_file, fheader, sizeof(lmjpeg_frameheader));
	if (ret<sizeof(lmjpeg_frameheader)) {
		printf("Error reading frame header! read %d bytes instead of %d\n",ret,sizeof(lmjpeg_frameheader));
		return -1;
	}

	if ((fheader->marker & LMJPEG_IO_FRAMEHEADERMASK)!=LMJPEG_IO_FRAMEHEADER) {
		printf("Error reading frame header! marker: %x frame number: %d frame siez: %d\n", fheader->marker, fheader->frame_num, fheader->size);
		return -1;
	}
	
	codec->frame_type=fheader->marker & ~LMJPEG_IO_FRAMEHEADERMASK;
	if ((fheader->marker & ~LMJPEG_IO_FRAMEHEADERMASK)==LMJPEG_IO_NFRAME) {
		return 0;
	}
	if (fheader->frame_num==0) {
		ret=lmjpeg_read_data(codec->lmjpeg_file, codec->config, sizeof(lmjpeg_fileheader));
		if (ret<sizeof(lmjpeg_fileheader)) {
			printf("Error reading! read %d bytes instead of %d\n",ret, sizeof(lmjpeg_fileheader));
			return -1;
		}
		if (codec->config->marker!=LMJPEG_IO_FILEHEADER) {
			printf("Error: read wrong file header!\n");
			return -1;
		}
		if (lmjpeg_setup_coder(codec)==-1) {
			printf("lmjpeg_setup_coder: something went wrong!\n");
		}
		size_tmp=sizeof(lmjpeg_fileheader);
	}
	
	ret=lmjpeg_read_data(codec->lmjpeg_file, img, fheader->size);	
	if (ret != fheader->size) {
		printf("Error reading! read %d bytes instead of %d\n",ret,fheader->size);
		return -1;
	}
	if ((size_tmp=size_chunk-sizeof(lmjpeg_frameheader)-size_tmp-fheader->size)!=0) {
		printf("skip %d bytes, pos: %d\n",size_tmp,(int) ftell(codec->lmjpeg_file));
		fseek(codec->lmjpeg_file, size_tmp, SEEK_CUR);
	}
	return ret;
}

/** close the avi file.
	@param codec a pointer to the CODEC struct
*/
void lmjpeg_reader_close(CODEC *codec) {
	fclose(codec->lmjpeg_file);
	codec->lmjpeg_file=NULL;
}

#endif
